﻿/**
 * Gesture Heatmaps Toolkit (.NET Framework 4.5 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 * 	    Lisa Anthony, Ph.D.
 *      CISE Department
 *      University of Florida
 *      Gainesville, FL 32611 USA
 *      lanthony@cise.ufl.edu
 *	
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for gesture heatmaps, and what should be 
 * used to cite it, is:
 *
 *	Vatavu, R.-D., Anthony, L. and Wobbrock, J.O. (2014).  
 *	  Gesture Heatmaps: Understanding Gesture Performance with Colorful Visualizations. 
 *	  In Proceedings of the 16th ACM on International Conference 
 *	  on Multimodal Interaction (ICMI '14). ACM, New York, NY, USA.
 *	  DOI=10.1145/2663204.2663256 http://dx.doi.org/10.1145/2663204.2663256
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2014, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor University of Florida, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/

using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Drawing;
using RelativeAccuracyMeasures;

namespace GHoST
{
    public partial class FormMain : Form
    {
        #region member data

        // gesture dataset 
        private Dictionary<string, GestureSet> _dataset = null; 

        // heatmap pre-computations (for fast display of gesture heatmaps)
        List<CentroidGesture> _centroidGestures = new List<CentroidGesture>();
        List<double[]> _centroidFeatureValues = new List<double[]>();
        List<double>[,] _confusionFeatureValues = null;
        List<double> _allFeatureValues = new List<double>();
        List<Bitmap> _bitmapsBackground = null;

        // heatmap computation settings
        private PointAlignmentType _pointAlignmentType = PointAlignmentType.Chronological;  // specifies how gesture points are aligned
        private FeatureType _featureType = FeatureType.None;                // the feature type being displayed by the heatmaps
        private ComputationType _computationType = ComputationType.None;    // the computation type (e.g., confusion matrix)
        private double _zoomLevel = 1.0;

        #endregion

        #region constructor, closing

        /// <summary>
        /// Constructs the main form.
        /// Performs initialization of the user interface components.
        /// </summary>
        public FormMain()
        {
            InitializeComponent();
            pbHeatmap.Image = new Bitmap(pbHeatmap.Width, pbHeatmap.Height);
            pbHeatmapScale.Image = new Bitmap(pbHeatmapScale.Width, pbHeatmapScale.Height);

            // initialize application settings
            folderBrowserDialog.SelectedPath = Application.StartupPath;

            // initialize UI heatmap settings
            panelHeatmap.Visible = false;
            trackBarStrokeThickness.Value = 15;  
            lblStrokeThickness.Text = trackBarStrokeThickness.Value.ToString() + " px";
            cbColorScheme.SelectedIndex = 0;
            chkShowBoxes.Checked = true;
            chkShowGestureType.Checked = true;
            chkShowSamplesInBackground.Checked = false;
            chkArrangeHeatmaps.Checked = false;
            chkManualOverrideScale.Checked = false;
            lblZoomLevel.Text = String.Format("Zoom heatmaps by {0:.0}", _zoomLevel);
            gboxAppearance.Enabled = false;

            // initialize menus
            computeToolStripMenuItem.Enabled = false;
            exportToolStripMenuItem.Enabled = false;
            alignPointsChronologicallyegUnistrokesToolStripMenuItem.PerformClick();
            shapeErrorToolStripMenuItem.PerformClick();

            lblFeatureType.Visible = false; 
        }

        /// <summary>
        /// Performs uninitialization actions when the form is closing.
        /// </summary>
        private void FormMain_FormClosing(object sender, FormClosingEventArgs e)
        {
            ReleaseBitmapsBackground();
        }

        /// <summary>
        /// Releases temporary bitmaps from the memory.
        /// </summary>
        private void ReleaseBitmapsBackground()
        {
            if (_bitmapsBackground != null)
            {
                for(int i = 0; i < _bitmapsBackground.Count; i++)
                    try
                    {
                        _bitmapsBackground[i].Dispose();
                        _bitmapsBackground[i] = null;
                    }
                    catch { }
                _bitmapsBackground = null;
            }
        }

        #endregion

        #region menu options (select feature type)

        private void absoluteTurningAngleToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            _featureType = FeatureType.AbsoluteTurningAngle;
            chkManualOverrideScale.Checked = false;

            lblFeatureType.Text = "Absolute turning angle";
            lblFeatureType.Left = panelHeatmap.Left + (panelHeatmap.Width - lblFeatureType.Width) / 2;
            lblFeatureType.Visible = true;

            absoluteTurningAngleToolStripMenuItem1.Checked = true;
            relativeTurningAngleToolStripMenuItem1.Checked = false;
            relativeLocalSpeedToolStripMenuItem1.Checked = false;
            localSpeedToolStripMenuItem.Checked = false;
            shapeErrorToolStripMenuItem.Checked = false;

            UpdateHeatmaps(true, true);
        }

        private void relativeTurningAngleToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            _featureType = FeatureType.RelativeTurningAngle;
            chkManualOverrideScale.Checked = false;

            lblFeatureType.Text = "Relative turning angle";
            lblFeatureType.Left = panelHeatmap.Left + (panelHeatmap.Width - lblFeatureType.Width) / 2;
            lblFeatureType.Visible = true;

            absoluteTurningAngleToolStripMenuItem1.Checked = false;
            relativeTurningAngleToolStripMenuItem1.Checked = true;
            relativeLocalSpeedToolStripMenuItem1.Checked = false;
            localSpeedToolStripMenuItem.Checked = false;
            shapeErrorToolStripMenuItem.Checked = false;

            UpdateHeatmaps(true, true);
        }

        private void localSpeedToolStripMenuItem_Click(object sender, EventArgs e)
        {
            _featureType = FeatureType.AbsoluteLocalSpeed;
            chkManualOverrideScale.Checked = false;

            lblFeatureType.Text = "Absolute local speed";
            lblFeatureType.Left = panelHeatmap.Left + (panelHeatmap.Width - lblFeatureType.Width) / 2;
            lblFeatureType.Visible = true;

            absoluteTurningAngleToolStripMenuItem1.Checked = false;
            relativeTurningAngleToolStripMenuItem1.Checked = false;
            relativeLocalSpeedToolStripMenuItem1.Checked = false;
            localSpeedToolStripMenuItem.Checked = true;
            shapeErrorToolStripMenuItem.Checked = false;

            UpdateHeatmaps(true, true);
        }

        private void relativeLocalSpeedToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            _featureType = FeatureType.RelativeLocalSpeed;
            chkManualOverrideScale.Checked = false;

            lblFeatureType.Text = "Relative local speed";
            lblFeatureType.Left = panelHeatmap.Left + (panelHeatmap.Width - lblFeatureType.Width) / 2;
            lblFeatureType.Visible = true;

            absoluteTurningAngleToolStripMenuItem1.Checked = false;
            relativeTurningAngleToolStripMenuItem1.Checked = false;
            relativeLocalSpeedToolStripMenuItem1.Checked = true;
            localSpeedToolStripMenuItem.Checked = false;
            shapeErrorToolStripMenuItem.Checked = false;

            UpdateHeatmaps(true, true);
        }

        private void shapeErrorToolStripMenuItem_Click(object sender, EventArgs e)
        {
            _featureType = FeatureType.RelativeShapeError;
            chkManualOverrideScale.Checked = false;

            lblFeatureType.Text = "Relative shape error";
            lblFeatureType.Left = panelHeatmap.Left + (panelHeatmap.Width - lblFeatureType.Width) / 2;
            lblFeatureType.Visible = true;

            absoluteTurningAngleToolStripMenuItem1.Checked = false;
            relativeTurningAngleToolStripMenuItem1.Checked = false;
            relativeLocalSpeedToolStripMenuItem1.Checked = false;
            localSpeedToolStripMenuItem.Checked = false;
            shapeErrorToolStripMenuItem.Checked = true;

            UpdateHeatmaps(true, true);
        }

        #endregion

        #region menu options (point alignment type)

        private void alignPointsChronologicallyegUnistrokesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            _pointAlignmentType = PointAlignmentType.Chronological;
            
            alignPointsChronologicallyegUnistrokesToolStripMenuItem.Checked = true;
            performPointcloudAlignmentegMultistrokesToolStripMenuItem.Checked = false;

            UpdateHeatmaps(true, true);
        }

        private void performPointcloudAlignmentegMultistrokesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            _pointAlignmentType = PointAlignmentType.PointCloudBestMatch;
            
            alignPointsChronologicallyegUnistrokesToolStripMenuItem.Checked = false;
            performPointcloudAlignmentegMultistrokesToolStripMenuItem.Checked = true;

            UpdateHeatmaps(true, true);
        }

        #endregion

        #region menu options (compute heatmaps)

        private void gestureHeatmapsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            _computationType = ComputationType.GestureHeatmaps;
            ComputeGestureHeatmaps(true, true);
        }

        private void chromaticConfusionMatrixToolStripMenuItem_Click(object sender, EventArgs e)
        {
            _computationType = ComputationType.ConfusionMatrix;
            ComputeConfusionMatrix(true, true);
        }

        #endregion

        #region heatmaps visual appearance options

        private void UpdateHeatmaps(bool recomputeHeatmaps, bool recomputeGesturesBackground)
        {
            switch (_computationType)
            {
                case ComputationType.GestureHeatmaps:
                    ComputeGestureHeatmaps(recomputeHeatmaps, recomputeGesturesBackground);
                    break;
                case ComputationType.ConfusionMatrix:
                    ComputeConfusionMatrix(recomputeHeatmaps, recomputeGesturesBackground);
                    break;
            }
        }

        private void RedrawGestureHeatmaps(object sender, EventArgs e)
        {
            UpdateHeatmaps(false, false);
        }

        private void RedrawGestureHeatmaps_ManualArrangement(object sender, EventArgs e)
        {
            UpdateHeatmaps(false, true);
        }

        private void trackBarStrokeThickness_Scroll(object sender, EventArgs e)
        {
            lblStrokeThickness.Text = trackBarStrokeThickness.Value.ToString() + " px";
        }

        private void trackBarStrokeThickness_MouseUp(object sender, MouseEventArgs e)
        {
            UpdateHeatmaps(false, false);
        }

        private void trackBarStrokeThickness_KeyUp(object sender, KeyEventArgs e)
        {
            UpdateHeatmaps(false, false);
        }

        private void chkManualArrangement_CheckedChanged(object sender, EventArgs e)
        {
            tbNumRows.Enabled = chkArrangeHeatmaps.Checked;
            if (!chkArrangeHeatmaps.Checked)
                UpdateHeatmaps(false, false);
        }

        private void btnZoomPlus_Click(object sender, EventArgs e)
        {
            _zoomLevel += 0.1;
            lblZoomLevel.Text = String.Format("Zoom heatmaps by {0:.0}", _zoomLevel);
            UpdateHeatmaps(false, true);
        }

        private void btnZoomMinus_Click(object sender, EventArgs e)
        {
            _zoomLevel = Math.Max(_zoomLevel - 0.1, 0.1);
            lblZoomLevel.Text = String.Format("Zoom heatmaps by {0:.0}", _zoomLevel);
            UpdateHeatmaps(false, true);
        }

        private void chkScale_CheckedChanged(object sender, EventArgs e)
        {
            tbMinScale.Enabled = chkManualOverrideScale.Checked;
            tbMaxScale.Enabled = chkManualOverrideScale.Checked;
            if (!chkManualOverrideScale.Checked)
                UpdateHeatmaps(false, false);
        }

        #endregion
    }
}
