﻿/**
 * Gesture Heatmaps Toolkit (.NET Framework 4.5 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 * 	    Lisa Anthony, Ph.D.
 *      CISE Department
 *      University of Florida
 *      Gainesville, FL 32611 USA
 *      lanthony@cise.ufl.edu
 *	
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for gesture heatmaps, and what should be 
 * used to cite it, is:
 *
 *	Vatavu, R.-D., Anthony, L. and Wobbrock, J.O. (2014).  
 *	  Gesture Heatmaps: Understanding Gesture Performance with Colorful Visualizations. 
 *	  In Proceedings of the 16th ACM on International Conference 
 *	  on Multimodal Interaction (ICMI '14). ACM, New York, NY, USA.
 *	  DOI=10.1145/2663204.2663256 http://dx.doi.org/10.1145/2663204.2663256
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2014, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor University of Florida, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/

using System.Windows.Forms;
using System.Collections.Generic;
using System.Drawing;
using RelativeAccuracyMeasures;

namespace GHoST
{
    public partial class FormMain
    {
        /// <summary>
        /// Computes the chromatic confusion matrix.
        /// </summary>
        private void ComputeConfusionMatrix(bool recomputeHeatmaps, bool recomputeGesturesBackground)
        {
            if (_dataset == null || _dataset.Count == 0)
                return;     // no data to process, return

            if (_featureType != FeatureType.RelativeShapeError)
            {
                MessageBox.Show("The confusion matrix can only be computed for the Shape Error feature.", "Info");
                return;
            }

            List<string> userSelectedGestureTypes = GetCheckedGestureTypes();
            WaitForm waitForm = null;

            using (Graphics graphics = Graphics.FromImage(pbHeatmapScale.Image))
                graphics.Clear(Color.White);
            pbHeatmapScale.Refresh();
            
            // compute heatmaps
            if (recomputeHeatmaps)
            {
                if (_pointAlignmentType == PointAlignmentType.PointCloudBestMatch)
                {
                    waitForm = new WaitForm("Please wait, the confusion matrix is being computed.\r\nDepending on the point alignment settings, this operation may take several seconds to complete.");
                    waitForm.Show();
                }
                ComputeConfusionMatrix(userSelectedGestureTypes, waitForm);
            }
            double minScale = 0, maxScale = 0;
            GetFeatureScale(out minScale, out maxScale);

            // the structure of the confusion matrix
            int numRows = userSelectedGestureTypes.Count + 1;
            int numCols = numRows;
            int spaceBetweenHeatmaps = 10;  // pixels
            int heatmapSize = 100; // pixels
            pbHeatmap.Image.Dispose();
            pbHeatmap.Image = new Bitmap(spaceBetweenHeatmaps + (spaceBetweenHeatmaps + heatmapSize) * numCols, spaceBetweenHeatmaps + (spaceBetweenHeatmaps + heatmapSize) * numCols);
            pbHeatmap.Width = pbHeatmap.Image.Width;
            pbHeatmap.Height = pbHeatmap.Image.Height;
            int leftMargin = spaceBetweenHeatmaps;
            int topMargin = spaceBetweenHeatmaps / 2;
            pbHeatmap.Location = new System.Drawing.Point(0, 0);

            if (waitForm != null)
                waitForm.Close();
            waitForm = new WaitForm("Please wait, the confusion matrix is being rendered.\r\nDepending on the visual appearance settings, this operation may take several seconds to complete.");
            waitForm.Show();

            // compute bitmaps background?
            if (recomputeGesturesBackground)
                ReleaseBitmapsBackground();
            if (chkShowSamplesInBackground.Checked)
            {
                if (_bitmapsBackground == null)
                {
                    _bitmapsBackground = new List<Bitmap>();
                    for (int i = 0; i < userSelectedGestureTypes.Count; i++)
                        _bitmapsBackground.Add(HeatmapGraphics.DrawGesturesBackground(heatmapSize, heatmapSize, _dataset[userSelectedGestureTypes[i]], _zoomLevel));
                }
            }

            // draw the confusion matrix
            ColorScheme colorScheme = GetDefaultColorSchemeForFeature(_featureType);
            using (Graphics graphics = Graphics.FromImage(pbHeatmap.Image))
            using (Font font = new System.Drawing.Font("Consolas", (int)(0.07 * heatmapSize)))
            {
                graphics.Clear(Color.White);

                // draw the first row and column of the matrix
                for (int i = 0; i < _centroidGestures.Count; i++)
                {
                    GestureSet heatmapSet = _dataset[userSelectedGestureTypes[i]];
                    Color[] blackColors = HeatmapGraphics.GetScaleColors(_confusionFeatureValues[i, i].ToArray(), minScale, maxScale, ColorScheme.Black);
                    using (Bitmap bitmap = HeatmapGraphics.DrawGestureHeatmap(heatmapSize, heatmapSize, _centroidGestures[i], blackColors, 2, heatmapSet, _zoomLevel))
                    {
                        graphics.DrawImage(bitmap, new System.Drawing.Rectangle(spaceBetweenHeatmaps / 2, spaceBetweenHeatmaps / 2 + (i + 1) * (heatmapSize + spaceBetweenHeatmaps), heatmapSize, heatmapSize));
                        graphics.DrawImage(bitmap, new System.Drawing.Rectangle(spaceBetweenHeatmaps / 2 + (i + 1) * (heatmapSize + spaceBetweenHeatmaps), spaceBetweenHeatmaps / 2, heatmapSize, heatmapSize));
                    }
                }
                graphics.DrawLine(Pens.LightGray, spaceBetweenHeatmaps / 2, heatmapSize + spaceBetweenHeatmaps, numCols * (heatmapSize + spaceBetweenHeatmaps), heatmapSize + spaceBetweenHeatmaps);
                graphics.DrawLine(Pens.LightGray, heatmapSize + spaceBetweenHeatmaps, spaceBetweenHeatmaps / 2, heatmapSize + spaceBetweenHeatmaps, numRows * (heatmapSize + spaceBetweenHeatmaps));

                // draw confusion matrix
                int indexProgress = 0;
                for (int i = 0; i < _centroidGestures.Count; i++)
                    for (int j = 0; j < _centroidGestures.Count; j++)
                    {
                        if (waitForm != null)
                        {
                            waitForm.IncrementProgress((int)((indexProgress + 1) * 100.0 / (_centroidGestures.Count * _centroidGestures.Count)));
                            Application.DoEvents();
                            indexProgress++;
                        }

                        System.Drawing.Rectangle rectangle = new System.Drawing.Rectangle(
                            spaceBetweenHeatmaps / 2 + (j + 1) * (heatmapSize + spaceBetweenHeatmaps), spaceBetweenHeatmaps / 2 + (i + 1) * (heatmapSize + spaceBetweenHeatmaps),
                            heatmapSize, heatmapSize
                        );

                        if (i == j)
                        {
                            using (SolidBrush grayBrush = new SolidBrush(Color.FromArgb(230, 230, 230)))
                                graphics.FillRectangle(grayBrush, rectangle);
                        }

                        if (chkShowSamplesInBackground.Checked)
                            graphics.DrawImage(_bitmapsBackground[j], rectangle);
                        
                        GestureSet heatmapSet = _dataset[userSelectedGestureTypes[j]];
                        Color[] colors = HeatmapGraphics.GetScaleColors(_confusionFeatureValues[i, j].ToArray(), minScale, maxScale, colorScheme);
                        using (Bitmap bitmap = HeatmapGraphics.DrawGestureHeatmap(heatmapSize, heatmapSize, _centroidGestures[i], colors, trackBarStrokeThickness.Value, heatmapSet, _zoomLevel))
                            graphics.DrawImage(bitmap, rectangle);

                        if (chkShowBoxes.Checked)
                            graphics.DrawRectangle(Pens.LightGray, rectangle);
                    }

            }
            pbHeatmap.Refresh();
            panelHeatmap.Visible = true;

            // draw heatmap scale
            int[] _histogram = ComputeHistogram(minScale, maxScale, _allFeatureValues, 100);
            using (Bitmap bitmap = HeatmapGraphics.DrawHeatmapScale(
                pbHeatmapScale.Image.Width, pbHeatmapScale.Image.Height,
                minScale, maxScale, colorScheme, _histogram))
            using (Graphics graphics = Graphics.FromImage(pbHeatmapScale.Image))
                graphics.DrawImage(bitmap, 0, 0);
            pbHeatmapScale.Refresh();
            pbHeatmapScale.Visible = true;

            if (waitForm != null)
                waitForm.Close();

            // enable menu options
            gboxAppearance.Enabled = true;
            exportToolStripMenuItem.Enabled = true;
            chkShowGestureType.Enabled = false;
            chkArrangeHeatmaps.Enabled = false;
        }

        /// <summary>
        /// Computes the confusion matrix.
        /// </summary>
        private void ComputeConfusionMatrix(List<string> userSelectedGestureTypes, WaitForm waitForm)
        {
            _centroidGestures = new List<CentroidGesture>();
            _confusionFeatureValues = new List<double>[userSelectedGestureTypes.Count, userSelectedGestureTypes.Count];
            _allFeatureValues = new List<double>();
                        
            int indexProgress = 0;
            for (int i = 0; i < userSelectedGestureTypes.Count; i++)
            {
                // compute the centroid gesture of a set
                GestureSet gestureArticulations = _dataset[userSelectedGestureTypes[i]];
                CentroidGesture centroid = new CentroidGesture(gestureArticulations.ToArray(), _pointAlignmentType, true);
                _centroidGestures.Add(centroid);

                // compute feature values for each point of the gesture centroid
                // with respect to all the other classes
                for (int j = 0; j < userSelectedGestureTypes.Count; j++)
                {
                    if (waitForm != null)
                    {
                        waitForm.IncrementProgress((int)((indexProgress + 1) * 100.0 / (userSelectedGestureTypes.Count * userSelectedGestureTypes.Count)));
                        Application.DoEvents();
                        indexProgress++;
                    }

                    double[] featureArray = ComputeFeatureValues(_featureType, _dataset[userSelectedGestureTypes[j]], centroid);
                    _confusionFeatureValues[i, j] = new List<double>();
                    _confusionFeatureValues[i, j].AddRange(featureArray);
                    _allFeatureValues.AddRange(featureArray);
                }
            }
        }
    }
}
