﻿/**
 * Gesture Heatmaps Toolkit (.NET Framework 4.5 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 * 	    Lisa Anthony, Ph.D.
 *      CISE Department
 *      University of Florida
 *      Gainesville, FL 32611 USA
 *      lanthony@cise.ufl.edu
 *	
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for gesture heatmaps, and what should be 
 * used to cite it, is:
 *
 *	Vatavu, R.-D., Anthony, L. and Wobbrock, J.O. (2014).  
 *	  Gesture Heatmaps: Understanding Gesture Performance with Colorful Visualizations. 
 *	  In Proceedings of the 16th ACM on International Conference 
 *	  on Multimodal Interaction (ICMI '14). ACM, New York, NY, USA.
 *	  DOI=10.1145/2663204.2663256 http://dx.doi.org/10.1145/2663204.2663256
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2014, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor University of Florida, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/

using System;
using System.Windows.Forms;
using System.Collections.Generic;
using System.Linq;
using System.Drawing;
using RelativeAccuracyMeasures;

namespace GHoST
{
    public partial class FormMain
    {
        /// <summary>
        /// Computes and displays colorful gesture heatmaps.
        /// </summary>
        private void ComputeGestureHeatmaps(bool recomputeHeatmaps, bool recomputeGesturesBackground)
        {
            if (_dataset == null || _dataset.Count == 0)
                return;     // no data to process, return

            List<string> userSelectedGestureTypes = GetCheckedGestureTypes();
            WaitForm waitForm = null;

            using (Graphics graphics = Graphics.FromImage(pbHeatmapScale.Image))
                graphics.Clear(Color.White);
            pbHeatmapScale.Refresh();

                        
            // compute heatmaps
            if (recomputeHeatmaps)
            {
                if (_pointAlignmentType == PointAlignmentType.PointCloudBestMatch)
                {
                    waitForm = new WaitForm("Please wait, heatmaps are being computed.\r\nDepending on the point alignment settings, this operation may take several seconds to complete.");
                    waitForm.Show();
                }
                ComputeGestureHeatmaps(userSelectedGestureTypes, waitForm);
            }
            double minScale = 0, maxScale = 0;
            GetFeatureScale(out minScale, out maxScale);

            // heatmaps are presented to the user as a matrix with numRows x numCols elements            
            int numRows = 4;
            if (chkArrangeHeatmaps.Checked)
            {
                try
                {
                    int value = int.Parse(tbNumRows.Text);
                    if (value <= 0 || value >= 4)
                        throw new Exception();
                    numRows = value;
                }
                catch { tbNumRows.Text = numRows.ToString(); }
            }
            else tbNumRows.Text = numRows.ToString();
            int numCols = (int)Math.Ceiling(userSelectedGestureTypes.Count * 1.0 / numRows);

            int spaceBetweenHeatmaps = 15;  // pixels
            int heatmapSize = 140 * 4 / numRows;  //pixels 
            pbHeatmap.Image.Dispose();
            pbHeatmap.Image = new Bitmap(spaceBetweenHeatmaps + (spaceBetweenHeatmaps + heatmapSize) * numCols, spaceBetweenHeatmaps + (spaceBetweenHeatmaps + heatmapSize) * numRows);
            pbHeatmap.Width = pbHeatmap.Image.Width;
            pbHeatmap.Height = pbHeatmap.Image.Height;
            int leftMargin = spaceBetweenHeatmaps;
            int topMargin = spaceBetweenHeatmaps / 2;
            pbHeatmap.Location = new System.Drawing.Point(Math.Max(0, (panelHeatmap.Width - pbHeatmap.Width) / 2), (panelHeatmap.Height - pbHeatmap.Height) / 2);

            // compute bitmaps for the gesture in the background?
            if (recomputeGesturesBackground)
                ReleaseBitmapsBackground();
            if (chkShowSamplesInBackground.Checked)
            {
                if (_bitmapsBackground == null)
                {
                    if (waitForm != null)
                        waitForm.Close();
                    waitForm = new WaitForm("Please wait, heatmaps are being rendered.\r\nDepending on the visual appearance settings, this operation may take several seconds to complete.");
                    waitForm.Show();

                    _bitmapsBackground = new List<Bitmap>();
                    for (int i = 0; i < userSelectedGestureTypes.Count; i++)
                    {
                        waitForm.IncrementProgress((int)((i + 1) * 100.0 / userSelectedGestureTypes.Count));
                        Application.DoEvents();
                        _bitmapsBackground.Add(HeatmapGraphics.DrawGesturesBackground(heatmapSize, heatmapSize, _dataset[userSelectedGestureTypes[i]], _zoomLevel));
                    }
                }
            }

            if (waitForm != null)
                waitForm.Close();
            waitForm = new WaitForm("Please wait, heatmaps are being rendered.\r\nDepending on the visual appearance settings, this operation may take several seconds to complete.");
            waitForm.Show();

            // draw heatmaps
            ColorScheme colorScheme = GetDefaultColorSchemeForFeature(_featureType);
            using (Graphics graphics = Graphics.FromImage(pbHeatmap.Image))
            using (Font font = new System.Drawing.Font("Consolas", (int)(0.07 * heatmapSize)))
            {
                graphics.Clear(Color.White);
                for (int i = 0; i < _centroidGestures.Count; i++)
                {
                    if (waitForm != null)
                    {
                        waitForm.IncrementProgress((int)((i + 1) * 100.0 / _centroidGestures.Count));
                        Application.DoEvents();
                    }

                    GestureSet heatmapSet = _dataset[userSelectedGestureTypes[i]];
                    Color[] colors = HeatmapGraphics.GetScaleColors(_centroidFeatureValues[i], minScale, maxScale, colorScheme);
                    using (Bitmap bitmap = HeatmapGraphics.DrawGestureHeatmap(heatmapSize, heatmapSize, _centroidGestures[i], colors, trackBarStrokeThickness.Value, heatmapSet, _zoomLevel))
                    {
                        int xHeatmap = leftMargin + (spaceBetweenHeatmaps + heatmapSize) * (i % numCols);
                        int yHeatmap = topMargin + (spaceBetweenHeatmaps + heatmapSize) * (i / numCols);
                        System.Drawing.Rectangle rectangle = new System.Drawing.Rectangle(xHeatmap, yHeatmap, heatmapSize, heatmapSize);

                        if (chkShowSamplesInBackground.Checked)
                            graphics.DrawImage(_bitmapsBackground[i], rectangle);
                        graphics.DrawImage(bitmap, rectangle);
                        if (chkShowBoxes.Checked)
                            graphics.DrawRectangle(Pens.LightGray, rectangle);
                        if (chkShowGestureType.Checked)
                        {
                            SizeF textSize = graphics.MeasureString(userSelectedGestureTypes[i], font);
                            graphics.FillRectangle(Brushes.White, xHeatmap + (heatmapSize - textSize.Width) / 2 - 2, yHeatmap + heatmapSize - textSize.Height / 2 - 2, textSize.Width + 4, textSize.Height + 4);
                            graphics.DrawString(userSelectedGestureTypes[i], font, Brushes.Black, xHeatmap + (heatmapSize - textSize.Width) / 2, yHeatmap + heatmapSize - textSize.Height / 2);
                        }
                    }
                }
            }
            pbHeatmap.Refresh();
            panelHeatmap.Visible = true;

            // draw heatmap scale
            int[] _histogram = ComputeHistogram(minScale, maxScale, _allFeatureValues, 100);
            using (Bitmap bitmap = HeatmapGraphics.DrawHeatmapScale(
                pbHeatmapScale.Image.Width, pbHeatmapScale.Image.Height,
                minScale, maxScale, colorScheme, _histogram))
            using (Graphics graphics = Graphics.FromImage(pbHeatmapScale.Image))
                graphics.DrawImage(bitmap, 0, 0);
            pbHeatmapScale.Refresh();
            pbHeatmapScale.Visible = true;

            if (waitForm != null)
                waitForm.Close();

            // gui menu
            gboxAppearance.Enabled = true;
            exportToolStripMenuItem.Enabled = true;
            chkShowGestureType.Enabled = true;
            chkArrangeHeatmaps.Enabled = true;
        }

        /// <summary>
        /// Computes gesture heatmaps.
        /// </summary>
        private void ComputeGestureHeatmaps(List<string> userSelectedGestureTypes, WaitForm waitForm)
        {
            _centroidGestures = new List<CentroidGesture>();
            _centroidFeatureValues = new List<double[]>();
            _allFeatureValues = new List<double>();

            for (int i = 0; i < userSelectedGestureTypes.Count; i++)
            {
                // give feedback to the user that computation may take a while
                if (waitForm != null)
                {
                    waitForm.IncrementProgress((int)((i + 1) * 100.0 / userSelectedGestureTypes.Count));
                    Application.DoEvents();
                }

                // compute centroid
                GestureSet gestureArticulations = _dataset[userSelectedGestureTypes[i]];
                CentroidGesture centroid = new CentroidGesture(gestureArticulations.ToArray(), _pointAlignmentType, true);
                _centroidGestures.Add(centroid);

                // compute feature values for each point of the gesture centroid
                double[] featureArray = ComputeFeatureValues(_featureType, gestureArticulations, centroid);
                _centroidFeatureValues.Add(featureArray);
                _allFeatureValues.AddRange(featureArray);
            }
        }

        /// <summary>
        /// Computes the scale for the the localized feature (auto or user-defined).
        /// </summary>
        private void GetFeatureScale(out double minScale, out double maxScale)
        {
            // compute the minimum and maximum values of the feature scale
            minScale = _allFeatureValues.Min();
            maxScale = _allFeatureValues.Max();
            if (chkManualOverrideScale.Checked)   // the user overrides the automatic scale
            {
                try
                {
                    double value = double.Parse(tbMinScale.Text);
                    minScale = value;
                }
                catch { tbMinScale.Text = String.Format("{0:.00}", minScale); }
                try
                {
                    double value = double.Parse(tbMaxScale.Text);
                    maxScale = value;
                }
                catch { tbMaxScale.Text = String.Format("{0:.00}", maxScale); }
            }
            else
            {
                tbMinScale.Text = String.Format("{0:.00}", minScale);
                tbMaxScale.Text = String.Format("{0:.00}", maxScale);
            }
        }

        /// <summary>
        /// Computes localized values for a gesture feature.
        /// </summary>
        private double[] ComputeFeatureValues(FeatureType featureType, GestureSet heatmapSet, CentroidGesture centroid)
        {
            List<double[]> allFeatures = new List<double[]>();

            foreach (Gesture gesture in heatmapSet)
            {
                double[] features = null;
                switch (featureType)
                {
                    case FeatureType.AbsoluteTurningAngle:
                        features = RelativeAccuracy.ComputeTurningAngleArray(gesture.Points);
                        break;
                    case FeatureType.AbsoluteLocalSpeed:
                        features = RelativeAccuracy.ComputeSpeedArray(gesture.Points);
                        break;
                    case FeatureType.RelativeTurningAngle:
                        RelativeAccuracy.BendingError(gesture, centroid, out features);
                        break;
                    case FeatureType.RelativeLocalSpeed:
                        RelativeAccuracy.VelocityError(gesture, centroid, out features);
                        break;
                    case FeatureType.RelativeShapeError:
                        RelativeAccuracy.ShapeError(gesture, centroid, out features);
                        break;
                }
                allFeatures.Add(features);
            }

            // compute mean
            double[] avg = new double[allFeatures[0].Length];
            Array.Clear(avg, 0, avg.Length);
            foreach (double[] features in allFeatures)
            {
                for (int i = 0; i < features.Length; i++)
                    avg[i] += features[i];
            }
            for (int i = 0; i < avg.Length; i++)
                avg[i] /= allFeatures.Count;
            return avg;
        }

        private int[] ComputeHistogram(double minScale, double maxScale, List<double> values, int numBins)
        {
            int[] histogram = new int[numBins];
            Array.Clear(histogram, 0, histogram.Length);

            if (Math.Abs(maxScale - minScale) > 1e-4)
            {
                for (int i = 0; i < values.Count; i++)
                {
                    double t = (values[i] - minScale) / (maxScale - minScale);
                    if (t < 0) t = 0;
                    if (t > 1) t = 1;
                    histogram[(int)(t * (histogram.Length - 1))]++;
                }
            }

            return histogram;
        }

        private List<string> GetCheckedGestureTypes()
        {
            List<string> selectedGestureTypes = new List<string>();
            foreach (KeyValuePair<string, GestureSet> record in _dataset)
                selectedGestureTypes.Add(record.Key);
            return selectedGestureTypes;
        }

        private ColorScheme GetDefaultColorSchemeForFeature(FeatureType featureType)
        {
            if (cbColorScheme.SelectedIndex == 0)   // default color map, which is dependent on the feature type
            {
                switch (featureType)
                {
                    case FeatureType.AbsoluteTurningAngle: return ColorScheme.Diverging_BlueToRed;
                    case FeatureType.RelativeTurningAngle: return ColorScheme.Sequential_Green;
                    case FeatureType.AbsoluteLocalSpeed: return ColorScheme.Sequential_Orange;
                    case FeatureType.RelativeLocalSpeed: return ColorScheme.Sequential_Green;
                    case FeatureType.RelativeShapeError: return ColorScheme.Rainbow;
                    default: return ColorScheme.None;
                }
            }
            else
            {
                switch (cbColorScheme.SelectedIndex)
                {
                    case 1: return ColorScheme.Diverging_BlueToRed;
                    case 2: return ColorScheme.Sequential_Orange;
                    case 3: return ColorScheme.Sequential_Green;
                    case 4: return ColorScheme.Rainbow;
                    default: return ColorScheme.Rainbow;
                }
            }
        }
    }
}
