﻿/**
 * Gesture Heatmaps Toolkit (.NET Framework 4.5 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 * 	    Lisa Anthony, Ph.D.
 *      CISE Department
 *      University of Florida
 *      Gainesville, FL 32611 USA
 *      lanthony@cise.ufl.edu
 *	
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for gesture heatmaps, and what should be 
 * used to cite it, is:
 *
 *	Vatavu, R.-D., Anthony, L. and Wobbrock, J.O. (2014).  
 *	  Gesture Heatmaps: Understanding Gesture Performance with Colorful Visualizations. 
 *	  In Proceedings of the 16th ACM on International Conference 
 *	  on Multimodal Interaction (ICMI '14). ACM, New York, NY, USA.
 *	  DOI=10.1145/2663204.2663256 http://dx.doi.org/10.1145/2663204.2663256
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2014, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor University of Florida, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/

using System;
using System.Windows.Forms;
using System.Collections.Generic;
using System.IO;
using RelativeAccuracyMeasures;

namespace GHoST
{
    public partial class FormMain
    {
        /// <summary>
        /// Loads gesture data for user-dependent analysis (one user).
        /// </summary>
        private void loadForUserdependentAnalysisToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (folderBrowserDialog.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                LoadGestureDataset(folderBrowserDialog.SelectedPath, true);
        }

        /// <summary>
        /// Loads gesture data for user-independent analysis (multiple users).
        /// </summary>
        private void loadForUserindependentAnalysisToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (folderBrowserDialog.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                LoadGestureDataset(folderBrowserDialog.SelectedPath, false);
        }

        /// <summary>
        /// Loads gesture data from XML files.
        /// </summary>
        private void LoadGestureDataset(string baseFolder, bool userDependentAnalysis)
        {
            using (WaitForm waitForm = new WaitForm("Please wait, gestures are loading.\r\nDepending on the size of the dataset, this operation may take several seconds to complete."))
            {
                waitForm.Show();

                // create new gesture dataset
                _dataset = new Dictionary<string, GestureSet>();

                // update UI
                lstGestureTypes.Items.Clear();
                tbNumParticipants.Text = "";
                tbNumGestureTypes.Text = "";
                tbNumGestureSamples.Text = "";
                chkShowSamplesInBackground.Checked = false;
                chkManualOverrideScale.Checked = false;
                gboxAppearance.Enabled = false;
                panelHeatmap.Visible = false;
                pbHeatmapScale.Visible = false;
                
                // number of folders = number of participants, one participant per folder
                string[] folders = userDependentAnalysis ? new string[] { baseFolder } : Directory.GetDirectories(baseFolder);

                // read gesture XML files from all folders
                int numGestureSamples = 0;
                for (int indexFolder = 0; indexFolder < folders.Length; indexFolder++)
                {
                    if (folders.Length > 1)
                    {
                        waitForm.IncrementProgress((int)(indexFolder * 100.0 / (folders.Length - 1)));
                        Application.DoEvents();
                    }

                    // number of XML files = number of gestures, one gesture sample per file
                    string[] xmlFiles = Directory.GetFiles(folders[indexFolder], "*.xml");
                    foreach (string xmlFile in xmlFiles)
                    {
                        Gesture gesture = GestureIO.ReadGesture(xmlFile);
                        string gestureType = gesture.Name.ToLower();

                        // add gesture to dataset
                        if (!_dataset.ContainsKey(gestureType))
                            _dataset.Add(gestureType, new GestureSet());
                        _dataset[gestureType].Add(gesture);

                        // count total number of gesture samples
                        numGestureSamples++;
                    }
                }

                // update UI: show details about the gesture dataset
                tbNumParticipants.Text = folders.Length.ToString();
                tbNumGestureTypes.Text = _dataset.Keys.Count.ToString();
                tbNumGestureSamples.Text = numGestureSamples.ToString();

                // update UI: show gesture types
                foreach (KeyValuePair<string, GestureSet> gestures in _dataset)
                {
                    lstGestureTypes.Items.Add(String.Format("{0} ({1})", gestures.Key, gestures.Value.Count));
                    lstGestureTypes.Items[lstGestureTypes.Items.Count - 1].Checked = true;
                }

                // update UI: enable menu options
                computeToolStripMenuItem.Enabled = true;
                exportToolStripMenuItem.Enabled = false;
            }
        }
    }
}
