﻿/**
 * Gesture Heatmaps Toolkit (.NET Framework 4.5 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 * 	    Lisa Anthony, Ph.D.
 *      CISE Department
 *      University of Florida
 *      Gainesville, FL 32611 USA
 *      lanthony@cise.ufl.edu
 *	
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for gesture heatmaps, and what should be 
 * used to cite it, is:
 *
 *	Vatavu, R.-D., Anthony, L. and Wobbrock, J.O. (2014).  
 *	  Gesture Heatmaps: Understanding Gesture Performance with Colorful Visualizations. 
 *	  In Proceedings of the 16th ACM on International Conference 
 *	  on Multimodal Interaction (ICMI '14). ACM, New York, NY, USA.
 *	  DOI=10.1145/2663204.2663256 http://dx.doi.org/10.1145/2663204.2663256
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2014, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor University of Florida, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using RelativeAccuracyMeasures;

namespace GHoST
{
    public class HeatmapGraphics
    {
        /// <summary>
        /// Produces a bitmap of width x height pixels that contains the heatmap visualization of a gesture
        /// colored with a specified color scheme (pointColors) and
        /// computed from a gesture set (set).
        /// </summary>
        public static Bitmap DrawGestureHeatmap(int width, int height, Gesture heatmap, Color[] pointColors, int strokeThickness, GestureSet set, double zoomLevel)
        {
            Bitmap bitmap = new Bitmap(width, height);
            double scaleFactor = zoomLevel * ComputeScaleFactor(set);
            System.Drawing.Rectangle drawingArea = new System.Drawing.Rectangle(0, 0, width, height);

            using (Graphics graphics = Graphics.FromImage(bitmap))
            {
                // render smooth, high-quality graphics 
                graphics.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.HighQualityBicubic;
                graphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;

                // scale heatmap points
                System.Drawing.Point[] points = new System.Drawing.Point[heatmap.Points.Length];
                for (int i = 0; i < heatmap.Points.Length; i++)
                    points[i] = ComputeBitmapPoint(heatmap.Points[i], drawingArea, scaleFactor);

                // draw border
                using (Pen pen = new Pen(Color.Black, strokeThickness))
                {
                    pen.EndCap = LineCap.Round;
                    for (int i = 0; i < points.Length - 1; i++)
                        if (heatmap.Points[i].StrokeID == heatmap.Points[i + 1].StrokeID)
                            graphics.DrawLine(pen, points[i], points[i + 1]);
                }

                // draw color path
                for (int i = 0; i < points.Length - 1; i++)
                    if (heatmap.Points[i].StrokeID == heatmap.Points[i + 1].StrokeID)
                        if (Math.Abs(points[i].X - points[i + 1].X) + Math.Abs(points[i].Y - points[i + 1].Y) >= 1) // safety check, otherwise LinearGradientBrush will throw an Exception
                        {
                            using (LinearGradientBrush brush = new LinearGradientBrush(points[i], points[i + 1], pointColors[i], pointColors[i + 1]))
                            using (Pen pen = new Pen(brush, strokeThickness / 2))
                            {
                                pen.EndCap = LineCap.Round;
                                graphics.DrawLine(pen, points[i], points[i + 1]);
                            }
                        }
            }

            return bitmap;
        }

        /// <summary>
        /// Draws a set of gestures on the background of the gesture heatmap.
        /// </summary>
        public static Bitmap DrawGesturesBackground(int width, int height, GestureSet set, double zoomLevel)
        {
            Bitmap bitmap = new Bitmap(width, height);
            double scaleFactor = zoomLevel * ComputeScaleFactor(set);
            System.Drawing.Rectangle drawingArea = new System.Drawing.Rectangle(0, 0, width, height);

            using (Graphics graphics = Graphics.FromImage(bitmap))
            {
                // render smooth, high-quality graphics 
                graphics.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.HighQualityBicubic;
                graphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;

                bool interpolateStrokeColors = true;
                Color firstColor = Color.Black, secondColor = Color.FromArgb(150, 150, 150);
                for (int j = 0; j < set.Count; j++)
                {
                    Gesture gesture = set[j];

                    // interpolate color between strokeColor and black
                    Color penColor = Color.Black;
                    if (interpolateStrokeColors)
                    {
                        double t = j * 1.0 / (set.Count - 1);
                        penColor = Color.FromArgb(
                            (int)((1 - t) * firstColor.R + t * secondColor.R),
                            (int)((1 - t) * firstColor.G + t * secondColor.G),
                            (int)((1 - t) * firstColor.B + t * secondColor.B)
                        );
                    }

                    // draw gesture path
                    using (Pen pen = new Pen(penColor, 2))
                    {
                        System.Drawing.Point previousPoint = ComputeBitmapPoint(gesture.Points[0], drawingArea, scaleFactor);
                        for (int i = 1; i < gesture.Points.Length; i++)
                        {
                            System.Drawing.Point currentPoint = ComputeBitmapPoint(gesture.Points[i], drawingArea, scaleFactor);
                            if (gesture.Points[i].StrokeID == gesture.Points[i - 1].StrokeID)
                                graphics.DrawLine(pen, previousPoint, currentPoint);
                            previousPoint = currentPoint;
                        }
                    }
                }
            }

            return bitmap;
        }

        #region heatmap color scale

        /// <summary>
        /// Draw a horizontal scale showing the correspondence between heat values and colors.
        /// </summary>
        public static Bitmap DrawHeatmapScale(int width, int height, double minScale, double maxScale, ColorScheme colorScheme, int[] bins)
        {
            Bitmap bitmap = new Bitmap(width, height);
            System.Drawing.Rectangle drawingArea = new System.Drawing.Rectangle((int)(0.05 * width), 10 + height / 2, (int)(0.9 * width), 10);

            List<Color> colors = null;
            switch (colorScheme)
            {
                case ColorScheme.Diverging_BlueToRed: colors = GetDivergingColors_BlueToRed(100); break;
                case ColorScheme.Sequential_Red: colors = GetSequentialColors_Red(100); break;
                case ColorScheme.Sequential_Orange: colors = GetSequentialColors_Orange(100); break;
                case ColorScheme.Sequential_Green: colors = GetSequentialColors_Green(100); break;
                case ColorScheme.Rainbow: colors = GetRainbowColors(100); break;
                case ColorScheme.Black: colors = GetUniqueColor(Color.Black, 100 * 6); break;
            }

            using (Graphics graphics = Graphics.FromImage(bitmap))
            {
                graphics.Clear(Color.White);

                // draw scale
                double barWidth = drawingArea.Width * 1.0 / (colors.Count - 1);
                for (int i = 0; i < colors.Count - 1; i++)
                {
                    using (LinearGradientBrush brush = new LinearGradientBrush(
                        new PointF(drawingArea.Left + (int)(i * barWidth) - 1, drawingArea.Top + drawingArea.Height / 2),
                        new PointF(drawingArea.Left + (int)((i + 1) * barWidth) + 1, drawingArea.Top + drawingArea.Height / 2),
                        colors[i], colors[i + 1]))
                    {
                        graphics.FillRectangle(brush, drawingArea.Left + (int)(i * barWidth), drawingArea.Top, (int)barWidth + 1, drawingArea.Height);
                    }
                }
                graphics.DrawRectangle(Pens.Black, drawingArea);

                // draw feature values (text)
                using (Font font = new Font("SansSerif", 12, FontStyle.Regular))
                {
                    for (double t = 0; t <= 1; t += 0.2)
                    {
                        string text = String.Format("{0:.00}", (1 - t) * minScale + t * maxScale);
                        SizeF sizeText = graphics.MeasureString(text, font);
                        graphics.DrawLine(Pens.Black, (int)(drawingArea.X + t * drawingArea.Width), drawingArea.Bottom - (int)(drawingArea.Height * 0.05), (int)(drawingArea.X + t * drawingArea.Width), drawingArea.Bottom + (int)(drawingArea.Height * 0.5));
                        graphics.DrawString(text, font, Brushes.Black, (int)(drawingArea.X + t * drawingArea.Width - sizeText.Width / 2), drawingArea.Bottom + (int)(drawingArea.Height * 0.55));
                    }
                }

                // draw histogram
                graphics.SmoothingMode = SmoothingMode.HighQuality;
               
                int max = 0;
                for (int i = 0; i < bins.Length; i++)
                    if (max < bins[i]) max = bins[i];
                if (max == 0) max = 1;
                System.Drawing.Point previousPoint = new System.Drawing.Point(drawingArea.Left, drawingArea.Top - 25);
                for (int i = 0; i < bins.Length; i++)
                {
                    System.Drawing.Point currentPoint = new System.Drawing.Point(
                        drawingArea.Left + drawingArea.Width * i / (bins.Length - 1), 
                        drawingArea.Top - 30 + (25 - bins[i] * 25 / max)
                    );
                    if (i > 0)
                    {
                        using(Pen pen = new Pen(Color.Black, 3))
                        using (SolidBrush brush = new SolidBrush(colors[i * (colors.Count - 1) / (bins.Length - 1)]))
                        {
                            graphics.DrawLine(pen, currentPoint, previousPoint);
                            graphics.FillPolygon(brush, new System.Drawing.Point[] 
                                { 
                                    new System.Drawing.Point(previousPoint.X, drawingArea.Top - 2),
                                    previousPoint,
                                    currentPoint,
                                    new System.Drawing.Point(currentPoint.X, drawingArea.Top - 2 )
                                }
                            );
                        }
                    }
                    previousPoint = currentPoint;
                }
                graphics.DrawRectangle(Pens.LightGray, drawingArea.Left, drawingArea.Top - 31, drawingArea.Width, 30);
            }

            return bitmap;
        }

        #endregion

        #region color schemes

        /// <summary>
        /// Computes the colors of a given color scheme.
        /// </summary>
        /// <returns></returns>
        public static Color[] GetScaleColors(double[] featureValues, double minScale, double maxScale, ColorScheme colorScheme, int resolution = 100)
        {
            List<Color> interpolatedColors = null;
            switch (colorScheme)
            {
                case ColorScheme.Diverging_BlueToRed: interpolatedColors = GetDivergingColors_BlueToRed(resolution); break;
                case ColorScheme.Sequential_Red: interpolatedColors = GetSequentialColors_Red(resolution); break;
                case ColorScheme.Sequential_Orange: interpolatedColors = GetSequentialColors_Orange(resolution); break;
                case ColorScheme.Sequential_Green: interpolatedColors = GetSequentialColors_Green(resolution); break;
                case ColorScheme.Rainbow: interpolatedColors = GetRainbowColors(resolution); break;
                case ColorScheme.Black: interpolatedColors = GetUniqueColor(Color.Black, resolution * 6); break;
            }

            Color[] colors = new Color[featureValues.Length];

            if (Math.Abs(maxScale - minScale) < 1e-4)
            {
                for (int i = 0; i < colors.Length; i++)
                    colors[i] = interpolatedColors[0];
            }
            else
            {
                for (int i = 0; i < featureValues.Length; i++)
                {
                    double t = (featureValues[i] - minScale) / (maxScale - minScale);
                    if (t < 0) t = 0.0;
                    if (t > 1) t = 1.0;
                    colors[i] = interpolatedColors[(int)(t * (interpolatedColors.Count - 1))];
                }
            }
            return colors;
        }

        private static List<Color> GetRainbowColors(int resolution)
        {
            Color[] baseColors = new Color[] { Color.Blue, Color.Green, Color.GreenYellow, Color.Yellow, Color.Orange, Color.DarkOrange, Color.Red };
            return GetInterpolatedColors(baseColors, resolution);
        }

        private static List<Color> GetDivergingColors_BlueToRed(int resolution)
        {
            Color[] baseColors = new Color[]    // from ColorBrewer, http://colorbrewer2.org/
            {
                Color.FromArgb(0x002166ac),
                Color.FromArgb(0x004393c3),
                Color.FromArgb(0x0092c5de),
                Color.FromArgb(0x00d1e5f0),
                Color.FromArgb(0x00f7f7f7),
                Color.FromArgb(0x00fddbc7),
                Color.FromArgb(0x00f4a582),
                Color.FromArgb(0x00d6604d),
                Color.FromArgb(0x00b2182b)
            };
            return GetInterpolatedColors(baseColors, resolution);
        }

        private static List<Color> GetSequentialColors_Red(int resolution)
        {
            Color[] baseColors = new Color[]
            {
                Color.FromArgb(0x00f7f7f7),
                Color.FromArgb(0x00fddbc7),
                Color.FromArgb(0x00f4a582),
                Color.FromArgb(0x00d6604d),
                Color.FromArgb(0x00b2182b)
            };
            return GetInterpolatedColors(baseColors, resolution);
        }

        private static List<Color> GetSequentialColors_Orange(int resolution)
        {
            Color[] baseColors = new Color[]
            {
                Color.FromArgb(0xffffcc),
                Color.FromArgb(0xffeda0),
                Color.FromArgb(0xfed976),
                Color.FromArgb(0xfeb24c),
                Color.FromArgb(0xfd8d3c),
                Color.FromArgb(0xfc4e2a),
                Color.FromArgb(0xe31a1c),
                Color.FromArgb(0xb10026)
            };
            return GetInterpolatedColors(baseColors, resolution);
        }

        private static List<Color> GetSequentialColors_Green(int resolution)
        {
            Color[] baseColors = new Color[]
            {
                Color.FromArgb(0xffffe5),
                Color.FromArgb(0xf7fcb9),
                Color.FromArgb(0xd9f0a3),
                Color.FromArgb(0xaddd8e),
                Color.FromArgb(0x78c679),
                Color.FromArgb(0x41ab5d),
                Color.FromArgb(0x238443),
                Color.FromArgb(0x005a32)
            };
            return GetInterpolatedColors(baseColors, resolution);
        }

        private static List<Color> GetInterpolatedColors(Color[] baseColors, int resolution)
        {
            List<Color> colors = new List<Color>();
            for (int i = 0; i < baseColors.Length - 1; i++)
                for (double s = 0.0; s <= 1.0; s += 1.0 / resolution)
                    colors.Add(Color.FromArgb(
                        (int)((1 - s) * baseColors[i].R + s * baseColors[i + 1].R),
                        (int)((1 - s) * baseColors[i].G + s * baseColors[i + 1].G),
                        (int)((1 - s) * baseColors[i].B + s * baseColors[i + 1].B)
                    ));
            return colors;
        }

        private static List<Color> GetUniqueColor(Color color, int resolution)
        {
            List<Color> colors = new List<Color>();
            for (int i = 0; i < resolution; i++)
                colors.Add(color);
            return colors;
        }

        #endregion

        #region conversions

        /// <summary>
        /// Computes and stores the bounding box of all gestures in the set.
        /// </summary>
        /// <param name="set"></param>
        private static double ComputeScaleFactor(GestureSet set)
        {
            RelativeAccuracyMeasures.Rectangle bboxOverall = new RelativeAccuracyMeasures.Rectangle(
                RelativeAccuracyMeasures.Point.AbsoluteMax,
                RelativeAccuracyMeasures.Point.AbsoluteMin
            );
         
            foreach (Gesture gesture in set)
            {
                RelativeAccuracyMeasures.Rectangle bbox = PointSet.BoundingBox(gesture.Points);
                if (bboxOverall.TopLeft.X > bbox.TopLeft.X) bboxOverall.TopLeft.X = bbox.TopLeft.X;
                if (bboxOverall.TopLeft.Y > bbox.TopLeft.Y) bboxOverall.TopLeft.Y = bbox.TopLeft.Y;
                if (bboxOverall.BottomRight.X < bbox.BottomRight.X) bboxOverall.BottomRight.X = bbox.BottomRight.X;
                if (bboxOverall.BottomRight.Y < bbox.BottomRight.Y) bboxOverall.BottomRight.Y = bbox.BottomRight.Y;
            }

            return 1.0 / Math.Max(bboxOverall.Width, bboxOverall.Height);
        }

        private static System.Drawing.Point ComputeBitmapPoint(
            RelativeAccuracyMeasures.Point point, 
            System.Drawing.Rectangle drawingArea,
            double scaleFactor)
        {
            int xMargin = (int)(0.1 * drawingArea.Width);
            int yMargin = (int)(0.1 * drawingArea.Height);
            return new System.Drawing.Point(
                (int)(xMargin + (point.X * scaleFactor + 0.5) * (drawingArea.Width - 2 * xMargin)),
                (int)(yMargin + (point.Y * scaleFactor + 0.5) * (drawingArea.Height - 2 * yMargin))
            );
        }

        #endregion
    }
}
